from . import REPR_INDENT

__all__ = ["StubBase", "Train", "Evaluate", "Lambda"]


class StubBase:
    """Base class for Stubs."""

    def __init__(self, *args, **kwargs):
        assert len(args) == 0, "There should not be any args only kwargs allowed."
        for k, v in kwargs.items():
            setattr(self, k, v)

    def replicate(self, copy: int = 1):
        # make copies of stubs with same kwargs
        copy = int(copy)
        assert copy > 0
        kwargs = self.__dict__
        copies = []
        for _ in range(copy):
            copies.append(self.__class__(**kwargs))
        return copies

    def __call__(self, copy: int = 1):
        if copy > 0:
            return self.replicate(copy)
        else:
            return None

    def __repr__(self):
        out = []
        out.append(self.__class__.__name__)
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            out.append(" " * REPR_INDENT + f"{k}: {v}")

        return "\n".join(out)


class Train(StubBase):
    """Train stub."""

    def __init__(self, dataloader: str, iteration: int = 0, **kwargs):

        assert isinstance(
            dataloader, str
        ), f"Dataloader should be string type but get type {type(dataloader)}"

        iteration = max(int(iteration), 0)

        kwargs["dataloader"] = dataloader
        kwargs["action"] = "train"
        kwargs["epoch"] = 1
        kwargs["iteration"] = iteration
        # additional options can be `optimizer`, `scheduler`

        super().__init__(**kwargs)


class Evaluate(StubBase):
    """Evaluation stub."""

    def __init__(self, dataloader: str, **kwargs):

        assert isinstance(
            dataloader, str
        ), f"Dataloader should be string type but get type {type(dataloader)}"
        kwargs["dataloader"] = dataloader
        kwargs["action"] = "evaluate"
        kwargs["epoch"] = 1
        # additional options can be metrics and eval specs

        super().__init__(**kwargs)


class Lambda(StubBase):
    """General action stub."""

    def __init__(self, action: str, **kwargs):

        assert isinstance(
            action, str
        ), f"Action should be string type but get type {type(action)}."

        assert action != "train", "Please use Train stub for training action."

        kwargs["action"] = action
        # e.g. save checkpoint, send emails, etc.

        super().__init__(**kwargs)



import inspect
from enum import Enum
from functools import partial
from typing import Callable, Dict, List, Optional

from .. import REPR_INDENT
from .base import EngineBase

__all__ = [
    "EventCategory",
    "EventHandler",
    "PreEpochHandler",
    "PostEpochHandler",
    "PreIterationHandler",
    "PostIterationHandler",
    "Engine",
]


class EventCategory(Enum):

    EPOCH_STARTS = "epoch_starts"
    EPOCH_FINISHES = "epoch_finishes"
    BEFORE_ITERATION = "before_iteration"
    AFTER_ITERATION = "after_iteration"


class EventHandler:
    """Base Class for Event Handlers."""

    category: EventCategory

    def __init__(
        self,
        action_function: Optional[Callable[[EngineBase], None]] = None,
        trigger_function: Optional[Callable[[EngineBase], bool]] = None,
        **kwargs
    ):
        self.action_function = action_function
        self.trigger_function = trigger_function

    def trigger(self, engine: EngineBase) -> bool:
        if self.trigger_function is None:
            return True

        return self.trigger_function(engine)

    def action(self, *args, **kwargs):
        if self.action_function is None:
            raise NotImplementedError(
                "Method `action` must be implemented if action_function is not provided."
            )
        else:
            self.action_function(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.trigger(*args, **kwargs):
            self.action(*args, **kwargs)

    def __repr__(self):
        lines = []
        lines.append(self.__class__.__name__)
        lines.append(" " * REPR_INDENT + "trigger function:")
        if self.trigger_function:
            for ln in inspect.getsource(self.trigger_function).split("\n"):
                lines.append(" " * REPR_INDENT * 2 + ln)
        lines.append(" " * REPR_INDENT + "action function:")
        if self.action_function:
            for ln in inspect.getsource(self.action_function).split("\n"):
                lines.append(" " * REPR_INDENT * 2 + ln)

        return "\n".join(lines)

    @classmethod
    def config(cls, *args, **kwargs):
        """
        Add additional config kwargs.

        For example
        ====================

        @EventHandler.config(param1=1.0, param2=2.0)
        def some_action_function(engine):
            ...
        """
        if args:
            raise ValueError("Only kwargs are allowed.")
        return partial(cls, **kwargs)


class PreEpochHandler(EventHandler):
    """Hanldes events when a new epoch starts."""

    category = EventCategory.EPOCH_STARTS

    def __init__(
        self,
        action_function: Callable[[EngineBase], None],
        trigger_function: Optional[Callable[[EngineBase], bool]] = None,
        every: int = 1,
    ):
        if trigger_function is None:
            trigger_function = lambda engine: engine.epoch % every == 0  # noqa
        super().__init__(action_function, trigger_function)


class PostEpochHandler(EventHandler):
    """Hanldes events when a new epoch finishes."""

    category = EventCategory.EPOCH_STARTS

    def __init__(
        self,
        action_function: Callable[[EngineBase], None],
        trigger_function: Optional[Callable[[EngineBase], bool]] = None,
        every: int = 1,
    ):
        if trigger_function is None:
            trigger_function = lambda engine: engine.epoch % every == 0  # noqa
        super().__init__(action_function, trigger_function)


class PreIterationHandler(EventHandler):
    """Handles events before each iteration."""

    category = EventCategory.BEFORE_ITERATION

    def __init__(
        self,
        action_function: Callable[[EngineBase], None],
        trigger_function: Optional[Callable[[EngineBase], bool]] = None,
        every: int = 1,
    ):
        if trigger_function is None:
            trigger_function = lambda engine: engine.iteration % every == 0  # noqa
        super().__init__(action_function, trigger_function)


class PostIterationHandler(EventHandler):
    """Handles events after each iteration."""

    category = EventCategory.AFTER_ITERATION

    def __init__(
        self,
        action_function: Callable[[EngineBase], None],
        trigger_function: Optional[Callable[[EngineBase], bool]] = None,
        every: int = 1,
    ):
        if trigger_function is None:
            trigger_function = lambda engine: engine.iteration % every == 0  # noqa
        super().__init__(action_function, trigger_function)


class Engine(EngineBase):
    """Engine with Event Handler plugin."""

    _event_handlers: Dict[EventCategory, List[EventHandler]]

    def __init__(self):
        super().__init__()
        self._event_handlers = {
            EventCategory.EPOCH_STARTS: [],
            EventCategory.EPOCH_FINISHES: [],
            EventCategory.BEFORE_ITERATION: [],
            EventCategory.AFTER_ITERATION: [],
        }

    def attach_event(self, handler: EventHandler):
        if isinstance(handler, EventHandler) and handler.category:
            self._event_handlers[handler.category].append(handler)
        else:
            raise TypeError("Category of handler must be specified.")

    def list_events(self, event_category: Optional[str] = None):
        return (
            self._event_handlers[EventCategory(event_category)]
            if event_category
            else self._event_handlers
        )

    def when_epoch_starts(self):
        for h in self._event_handlers[EventCategory.EPOCH_STARTS]:
            h(self)

    def when_epoch_finishes(self):
        for h in self._event_handlers[EventCategory.EPOCH_FINISHES]:
            h(self)

    def before_iteration(self):
        for h in self._event_handlers[EventCategory.BEFORE_ITERATION]:
            h(self)

    def after_iteration(self):
        for h in self._event_handlers[EventCategory.AFTER_ITERATION]:
            h(self)



import inspect
from functools import partial
from typing import Callable

from .buffer import BufferBase, ExponentialMovingAverage, to_buffer
from .events import Engine

__all__ = ["Automated"]


class Automated(Engine):
    """
    Automated Engine Given a forward generator function, `from_forward` will
    return an Automated engine class.

    For example
    ==================

    import torch
    import torch.nn as nn
    import torch.nn.functional as F

    @Automated.config(smooth_window=100)
    def classification(engine, batch):
        # the first arg must be a place holder for engine class

        engine.train()
        x, y = batch
        lgs = engine.model(x)
        loss = F.cross_entropy(lgs, y)

        yield "loss", loss.item()
        # metrics will be registered as buffers

        acc = (lgs.max(-1).indices == y).float().mean()

        yield "acc", acc.item()

    # or alternatively
    # `from_forward` will be deprecated
    classification = Automated.from_forward(classification)

    # attach other components such as model, optimizer, dataloader, etc.
    eng.attach(model=nn.Linear(2, 2))
    ...
    """

    def __init__(
        self,
        core_function: Callable,
        alpha: float = 1e-2,
        smooth_window: int = 50,
        buffer_type: BufferBase = ExponentialMovingAverage,
        **kwargs
    ):
        assert inspect.isgeneratorfunction(core_function), (
            "The forward function must be a generator function "
            "with first arg being engine class placeholder."
        )
        super().__init__()
        self.core = partial(core_function, self)
        buffer_names = _find_outputs(core_function)
        self.attach(**{n: buffer_type(alpha, smooth_window) for n in buffer_names})

    @classmethod
    def config(cls, **kwargs):
        """
        Used as decorator for core function allowing user to attach additional
        init keyword args Examples.

        @Automated
        def core_func(ng, batch):
            ...


        @Automated.config(smooth_window=100)
        def core_func(ng, batch):
            ...
        """
        return partial(cls, **kwargs)

    def core(self, batch, **kwargs):
        raise NotImplementedError("Method `core` must be implemented.")

    @to_buffer("buffer_registry")
    def per_batch(self, batch, **kwargs):
        return self.core(batch, **kwargs)

    def attach(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def from_forward(cls, func, smooth_window=50, **kwargs):
        """
        This method is deprecated.

        Use init as decorator or cls.config(...) as decorator
        """
        eng = cls(func, smooth_window=smooth_window, **kwargs)
        return eng


def _find_outputs(func):
    source = inspect.getsource(func)
    names = []
    for line in source.splitlines():
        line = line.strip()
        if line.startswith("yield"):
            line = line.replace("yield", "")
            names.append(line.split(",")[0].strip()[1:-1])
    return names
